import importlib
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Type, TypeVar

from ..file_io import LoadsFromYamlFile, SavesToYamlFile
from ..pipeline import Step
from ..schema.schema import (
    AggregatedIntrospectiveIngestionComponent,
    GraphSchema,
    IntrospectiveIngestionComponent,
)
from .pipeline_definition import PipelineDefinition
from .pipeline_scope import PipelineScope
from .run_request import RunRequest

T = TypeVar("T", bound=Step)


class Project(
    AggregatedIntrospectiveIngestionComponent, LoadsFromYamlFile, SavesToYamlFile
):
    """A `Project` represents a collection of pipelines."""

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                Optional("scopes"): {
                    str: PipelineScope.describe_yaml_schema(),
                },
                Optional("imports"): [str],
            }
        )

    @classmethod
    def from_file_data(cls, data) -> "Project":
        scopes_data = data.pop("scopes", {})
        scopes = [
            PipelineScope.from_file_data(*scope_data)
            for scope_data in scopes_data.items()
        ]
        imports = data.pop("imports", [])
        return cls(scopes, imports)

    def to_file_data(self):
        return {
            "scopes": {
                scope.name: scope.to_file_data()
                for scope in self.scopes_by_name.values()
            },
            "imports": self.imports,
        }

    def __init__(self, scopes: List[PipelineScope], imports: List[str]):
        self.scopes_by_name: Dict[str, PipelineScope] = {}
        for scope in scopes:
            self.add_scope(scope)
        self.imports = imports

    def ensure_modules_are_imported(self):
        for module in self.imports:
            importlib.import_module(module)

    async def run(self, request: RunRequest):
        for scope in self.scopes_by_name.values():
            await scope.run_request(request)

    def add_scope(self, scope: PipelineScope):
        self.scopes_by_name[scope.name] = scope

    def get_scopes_by_name(self, scope_name: Optional[str]) -> Iterable[PipelineScope]:
        if scope_name is None:
            return self.scopes_by_name.values()

        if scope_name not in self.scopes_by_name:
            return []

        return [self.scopes_by_name[scope_name]]

    def delete_pipeline(
        self,
        scope_name: Optional[str],
        pipeline_name: str,
        remove_pipeline_file: bool = True,
        missing_ok: bool = True,
    ):
        for scopes in self.get_scopes_by_name(scope_name):
            scopes.delete_pipeline(
                pipeline_name,
                remove_pipeline_file=remove_pipeline_file,
                missing_ok=missing_ok,
            )

    def get_schema(self, type_overrides_file: Optional[Path] = None) -> GraphSchema:
        schema = self.generate_graph_schema()
        if type_overrides_file is not None:
            schema.apply_type_overrides_from_file(type_overrides_file)
        return schema

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return self.scopes_by_name.values()

    def dig_for_step_of_type(
        self, step_type: Type[T]
    ) -> Iterable[Tuple[PipelineDefinition, int, T]]:
        for scope in self.scopes_by_name.values():
            for pipeline_definition in scope.pipelines_by_name.values():
                pipeline_steps = (
                    pipeline_definition.initialize_for_introspection().steps
                )
                for idx, step in enumerate(pipeline_steps):
                    if isinstance(step, step_type):
                        yield pipeline_definition, idx, step



from pathlib import Path
from typing import List

from ...project import PipelineDefinition, PipelineScope, Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class GenerateProject(Operation):
    def __init__(
        self,
        project_root: Path,
        pipelines: List[Path],
        source_modules: List[Path],
        database: str,
    ) -> None:
        self.pipelines = pipelines
        self.source_modules = source_modules
        self.project_root = project_root
        self.database = database

    async def perform(self, _: NodestreamCommand):
        imports = self.generate_import_directives()
        scope = self.generate_pipeline_scope()
        return Project([scope], imports)

    def generate_import_directives(self) -> List[str]:
        project_imports = [
            str(path.relative_to(self.project_root).with_suffix(""))
            .replace("/", ".")
            .replace("\\", ".")
            for path in self.source_modules
            if "__init__" not in path.name
        ]
        project_imports.append(f"nodestream.databases.{self.database}")
        return project_imports

    def generate_pipeline_scope(self):
        pipeline_definitions = [
            PipelineDefinition.from_path(path.relative_to(self.project_root))
            for path in self.pipelines
        ]
        return PipelineScope("default", pipeline_definitions)



from dataclasses import dataclass

from ..pipeline import PipelineInitializationArguments
from ..pipeline.meta import start_context
from .pipeline_definition import PipelineDefinition
from .pipeline_progress_reporter import PipelineProgressReporter


@dataclass
class RunRequest:
    """A `RunRequest` represents a request to run a pipeline."""

    pipeline_name: str
    initialization_arguments: PipelineInitializationArguments
    progress_reporter: PipelineProgressReporter

    async def execute_with_definition(self, definition: PipelineDefinition):
        with start_context(self.pipeline_name):
            pipeline = definition.initialize(self.initialization_arguments)
            await self.progress_reporter.execute_with_reporting(pipeline)



from ...project.audits import TTLAudit
from .audit_command import AuditCommand


class AuditTTLs(AuditCommand):
    name = "audit ttls"
    description = "Audit the project for missing TTLs"
    audit = TTLAudit



from pathlib import Path
from typing import List

from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

PROJECT_MODULES_TO_CREATE = [
    {"module_name": "__init__"},
    {
        "module_name": "argument_resolvers",
        "import_statement": "from nodestream.pipeline.argument_resolvers import ArgumentResolver",
        "guide_reference": "# See: https://nodestream-proj.github.io/nodestream/extending-nodestream/creating-your-own-argument-resolver",
    },
    {
        "module_name": "normalizers",
        "import_statement": "from nodestream.pipeline.normalizers import Normalizer",
        "guide_reference": "# See: https://nodestream-proj.github.io/nodestream/extending-nodestream/creating-your-own-normalizer",
    },
    {
        "module_name": "value_providers",
        "import_statement": "from nodestream.pipeline.value_providers import ValueProvider",
        "guide_reference": "# See: https://nodestream-proj.github.io/nodestream/extending-nodestream/creating-your-own-value-provider",
    },
]


class GeneratePythonScaffold(Operation):
    def __init__(self, project_root: Path) -> None:
        self.project_root = project_root

    async def perform(self, _: NodestreamCommand) -> List[Path]:
        return [
            self.write_module(module_configuration)
            for module_configuration in PROJECT_MODULES_TO_CREATE
        ]

    def get_file_path(self, module_name) -> Path:
        return (
            self.project_root / self.project_root.absolute().name / f"{module_name}.py"
        )

    def write_module(self, module_config) -> Path:
        path = self.get_file_path(module_config["module_name"])
        self.ensure_path_to_file_exists(path)
        self.write_module_contents_at_path(path, module_config)
        return path

    def write_module_contents_at_path(self, path: Path, module_config):
        lines = []
        if import_statement := module_config.get("import_statement"):
            lines.append(import_statement)
            lines.append("\n")

        if guide_reference := module_config.get("guide_reference"):
            lines.append(guide_reference)

        with open(path, "w") as fp:
            fp.writelines(lines)

    def ensure_path_to_file_exists(self, path: Path):
        path.parent.mkdir(parents=True, exist_ok=True)



from .new import New
from .nodestream_command import NodestreamCommand
from .remove import Remove
from .run import Run
from .scaffold import Scaffold
from .show import Show

__all__ = ("New", "NodestreamCommand", "Remove", "Run", "Scaffold", "Show")



import os.path
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict

from ..file_io import LoadsFromYaml, SavesToYaml
from ..pipeline import Pipeline, PipelineFileLoader, PipelineInitializationArguments
from ..schema.schema import IntrospectiveIngestionComponent


def get_default_name(file_path: Path) -> str:
    return file_path.stem


@dataclass
class PipelineDefinition(IntrospectiveIngestionComponent, SavesToYaml, LoadsFromYaml):
    """A `PipelineDefinition` represents a pipeline that can be loaded from a file."""

    name: str
    file_path: Path
    annotations: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_path(cls, file_path: Path):
        return cls(get_default_name(file_path), file_path)

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Or, Schema

        return Schema(
            Or(
                str,
                {
                    Optional("name"): str,
                    "path": os.path.exists,
                    Optional("annotations"): {str: Or(str, int, float, bool)},
                },
            )
        )

    @classmethod
    def from_file_data(cls, data, parent_annotations):
        if isinstance(data, str):
            data = {"path": data}

        file_path = Path(data.pop("path"))
        name = data.pop("name", get_default_name(file_path))
        annotations = data.pop("annotations", {})
        return cls(name, file_path, {**parent_annotations, **annotations})

    def to_file_data(self):
        using_default_name = self.name == self.file_path.stem
        if using_default_name and not self.annotations:
            return str(self.file_path)

        result = {"path": str(self.file_path)}
        if not using_default_name:
            result["name"] = self.name
        if self.annotations:
            result["annotations"] = self.annotations

        return result

    def initialize(self, init_args: PipelineInitializationArguments) -> Pipeline:
        return PipelineFileLoader(self.file_path).load_pipeline(init_args)

    def remove_file(self, missing_ok: bool = True):
        self.file_path.unlink(missing_ok=missing_ok)

    def initialize_for_introspection(self) -> Pipeline:
        return self.initialize(PipelineInitializationArguments.for_introspection())

    def gather_object_shapes(self):
        return self.initialize_for_introspection().gather_object_shapes()

    def gather_present_relationships(self):
        return self.initialize_for_introspection().gather_present_relationships()

    def gather_used_indexes(self):
        return self.initialize_for_introspection().gather_used_indexes()



import json
from abc import ABC, abstractmethod
from typing import Any, Iterable

from ....model import JsonLikeDocument
from ....subclass_registry import SubclassRegistry
from ...flush import Flush
from ..extractor import Extractor

STREAM_CONNECTOR_SUBCLASS_REGISTRY = SubclassRegistry()
STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY = SubclassRegistry()

DEFAULT_TIMEOUT = 60
DEFAULT_MAX_RECORDS = 100


@STREAM_CONNECTOR_SUBCLASS_REGISTRY.connect_baseclass
class StreamConnector(ABC):
    @abstractmethod
    async def connect(self):
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self):
        raise NotImplementedError

    @abstractmethod
    async def poll(self, timeout: int, max_records: int) -> Iterable[Any]:
        raise NotImplementedError


@STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY.connect_baseclass
class StreamRecordFormat(ABC):
    @abstractmethod
    def parse(self, record: Any) -> JsonLikeDocument:
        raise NotImplementedError


class JsonStreamRecordFormat(StreamRecordFormat, alias="json"):
    def parse(self, record: Any) -> JsonLikeDocument:
        return json.loads(record)


class StreamExtractor(Extractor):
    """A StreamExtractor implements the standard behavior of polling data from a stream.

    The StreamExtractor requires both a StreamConnector and a StreamRecordFormat to delegate
    to for the actual polling implementation and parsing of the data, respectively.
    """

    @classmethod
    def from_file_data(
        cls,
        connector: str,
        record_format: str,
        timeout: int = DEFAULT_TIMEOUT,
        max_records: int = DEFAULT_MAX_RECORDS,
        **connector_args
    ):
        object_format_cls = STREAM_OBJECT_FORMAT_SUBCLASS_REGISTRY.get(record_format)
        connector_cls = STREAM_CONNECTOR_SUBCLASS_REGISTRY.get(connector)
        return cls(
            timeout=timeout,
            max_records=max_records,
            record_format=object_format_cls(),
            connector=connector_cls(**connector_args),
        )

    def __init__(
        self,
        connector: StreamConnector,
        record_format: StreamRecordFormat,
        timeout: int,
        max_records: int,
    ):
        self.connector = connector
        self.record_format = record_format
        self.timeout = timeout
        self.max_records = max_records

    async def poll(self):
        return await self.connector.poll(
            timeout=self.timeout, max_records=self.max_records
        )

    async def extract_records(self):
        await self.connector.connect()
        try:
            results = tuple(await self.poll())
            if len(results) == 0:
                yield Flush
            else:
                for record in results:
                    yield self.record_format.parse(record)
        finally:
            await self.connector.disconnect()



from ..project import Project
from .audit_printer import AuditPrinter


class Audit:
    def __init__(self, printer: AuditPrinter) -> None:
        self.printer = printer
        self.success_count = 0
        self.warning_count = 0
        self.failure_count = 0

    async def run(self, project: Project) -> None:
        pass

    def success(self, message: str) -> None:
        self.success_count += 1
        self.printer.print_success(message)

    def failure(self, message: str) -> None:
        self.failure_count += 1
        self.printer.print_failure(message)

    def warning(self, message: str) -> None:
        self.warning_count += 1
        self.printer.print_warning(message)
